@@ -8,27 +8,49 @@
 //
 // SPDX-License-Identifier: MPL-2.0
 
+//! Implements systemd's [Boot Loader Specification](https://systemd.io/BOOT_LOADER_SPECIFICATION/)
+//!
+//! This API can parse and modify a Boot Loader Spec entry file.
+//! It supports the [Fedora/GRUB specific commands](https://fedoraproject.org/wiki/Changes/BootLoaderSpecByDefault).
+//!
+//! This library can be used in a no_std environment that
+//! supports dynamic memory allocation by disabling
+//! the "std" feature using --no-default-features.
+//!
+//! NOTE: At the moment, if you parse a BLSEntry with full-line
+//! comments and write it back, all comment lines will be consolidated
+//! in the header.
 
 #![cfg_attr(not(feature = "std"), no_std)]
-
 #![cfg(not(feature = "std"))]
 extern crate alloc;
 
+#[cfg(not(feature = "std"))]
+use alloc::format;
 #[cfg(not(feature = "std"))]
 use alloc::string::String;
 #[cfg(not(feature = "std"))]
 use alloc::vec::Vec;
-#[cfg(not(feature = "std"))]
-use alloc::format;
 
-pub type KeyValue = Vec<String>;
-
-#[derive(Debug)]
+/// BLSValue generalizes a value that may have an inline comment
+///
+/// - ```BLSValue::Value(String)```: the string represent the argument string for the BLS command
+/// - ```BLSValue::ValueWithComment(String, String)```: holds string that represents the BLS command argument as well as a comment string at the end of the line
+#[derive(Debug, PartialEq)]
 pub enum BLSValue {
     Value(String),
     ValueWithComment(String, String),
 }
 
+#[derive(Debug, PartialEq)]
+pub enum ValueSetPolicy {
+    ReplaceAll,
+    Append,
+    Prepend,
+    InsertAt(usize),
+}
+
+/// BLSEntry represents the contents of a BLS entry file
 #[derive(Debug)]
 pub struct BLSEntry {
     pub title: Option<BLSValue>,
@@ -51,6 +73,7 @@ pub struct BLSEntry {
 }
 
 impl BLSEntry {
+    /// Allocates a new instance of BLSEntry, all optional members are initialized to None and ```linux``` is set with an empty string
     pub fn new() -> BLSEntry {
         BLSEntry {
             title: None,
@@ -72,6 +95,9 @@ impl BLSEntry {
         }
     }
 
+    /// Parses a Boot Loader Spec entry UTF-8 buffer, returns a BLSEntry instance if successful, an error String if there was an error
+    /// Note that any comment lines that are then rendered using BLSEntry::parse() will be pushed to the header of the file as the
+    /// order commands and comments are not preserved.
     pub fn parse(buffer: &str) -> Result<BLSEntry, String> {
         let mut entry = BLSEntry::new();
         let mut has_kernel = false;
@@ -93,7 +119,12 @@ impl BLSEntry {
                 if key_value[0] == "linux" {
                     has_kernel = true;
                 }
-                if let Err(error) = entry.set(key_value[0], String::from(key_value[1]), comment) {
+                if let Err(error) = entry.set(
+                    key_value[0],
+                    String::from(key_value[1]),
+                    comment,
+                    ValueSetPolicy::Append,
+                ) {
                     return Err(error);
                 }
             } else {
@@ -113,6 +144,7 @@ impl BLSEntry {
         }
     }
 
+    /// Renders the BLSEntry content into a UTF-8 String
     pub fn render(&self) -> String {
         let mut content = String::new();
 
@@ -171,13 +203,33 @@ impl BLSEntry {
         content
     }
 
-    pub fn set(&mut self, key: &str, value: String, comment: Option<String>) -> Result<(), String> {
+    /// Sets a value
+    pub fn set(
+        &mut self,
+        key: &str,
+        value: String,
+        comment: Option<String>,
+        set_policy: ValueSetPolicy,
+    ) -> Result<(), String> {
         fn value_generator(value: String, comment: Option<String>) -> BLSValue {
             match comment {
                 Some(comment) => BLSValue::ValueWithComment(value, comment),
                 None => BLSValue::Value(value),
             }
         }
+
+        fn push_value(values: &mut Vec<BLSValue>, val: BLSValue, policy: ValueSetPolicy) {
+            match policy {
+                ValueSetPolicy::Append => values.push(val),
+                ValueSetPolicy::InsertAt(i) => values.insert(i, val),
+                ValueSetPolicy::Prepend => values.insert(0, val),
+                ValueSetPolicy::ReplaceAll => {
+                    values.clear();
+                    values.push(val);
+                }
+            }
+        }
+
         match key {
             "title" => self.title = Some(value_generator(value, comment)),
             "version" => self.version = Some(value_generator(value, comment)),
@@ -185,19 +237,28 @@ impl BLSEntry {
             "sort-key" => self.sort_key = Some(value_generator(value, comment)),
             "linux" => self.linux = value_generator(value, comment),
             "efi" => self.efi = Some(value_generator(value, comment)),
-            "initrd" => self.initrd.push(value_generator(value, comment)),
-            "options" => self.options.push(value_generator(value, comment)),
             "devicetree" => self.devicetree = Some(value_generator(value, comment)),
             "devicetree-overlay" => self.devicetree_overlay = Some(value_generator(value, comment)),
             "architecture" => self.architecture = Some(value_generator(value, comment)),
             "grub_hotkey" => self.grub_hotkey = Some(value_generator(value, comment)),
             "grub_users" => self.grub_users = Some(value_generator(value, comment)),
-            "grub_class" => {
-                self.grub_class.push(value_generator(value, comment));
-            }
-            "grub_arg" => {
-                self.grub_arg = Some(value_generator(value, comment));
-            }
+            "grub_arg" => self.grub_arg = Some(value_generator(value, comment)),
+
+            "initrd" => push_value(
+                &mut self.initrd,
+                value_generator(value, comment),
+                set_policy,
+            ),
+            "options" => push_value(
+                &mut self.options,
+                value_generator(value, comment),
+                set_policy,
+            ),
+            "grub_class" => push_value(
+                &mut self.grub_class,
+                value_generator(value, comment),
+                set_policy,
+            ),
             _ => return Err(format!("Invalid key {}", key)),
         }
 
@@ -207,8 +268,14 @@ impl BLSEntry {
 
 #[cfg(test)]
 mod bls_tests {
+    use super::String;
+
+    #[cfg(not(feature = "std"))]
+    use alloc::vec;
+
     use super::BLSEntry;
     use super::BLSValue;
+    use super::ValueSetPolicy;
 
     #[test]
     fn new_entry() {
@@ -265,4 +332,63 @@ mod bls_tests {
         let entry = BLSEntry::parse(entry_txt);
         assert!(entry.is_err());
     }
+
+    #[test]
+    fn set_value_policies() {
+        // Append
+        let mut entry = BLSEntry::new();
+        let _ = entry.set("options", String::from("foo"), None, ValueSetPolicy::Append);
+        let _ = entry.set("options", String::from("bar"), None, ValueSetPolicy::Append);
+        let _ = entry.set("options", String::from("baz"), None, ValueSetPolicy::Append);
+
+        assert_eq!(
+            entry.options,
+            vec![
+                BLSValue::Value(String::from("foo")),
+                BLSValue::Value(String::from("bar")),
+                BLSValue::Value(String::from("baz"))
+            ]
+        );
+
+        // InsertAt
+        let _ = entry.set(
+            "options",
+            String::from("lol"),
+            None,
+            ValueSetPolicy::InsertAt(1),
+        );
+        assert_eq!(
+            entry.options,
+            vec![
+                BLSValue::Value(String::from("foo")),
+                BLSValue::Value(String::from("lol")),
+                BLSValue::Value(String::from("bar")),
+                BLSValue::Value(String::from("baz"))
+            ]
+        );
+
+        // ReplaceAll
+        let _ = entry.set(
+            "options",
+            String::from("wtf"),
+            None,
+            ValueSetPolicy::ReplaceAll,
+        );
+        assert_eq!(entry.options, vec![BLSValue::Value(String::from("wtf"))]);
+
+        // Prepend
+        let _ = entry.set(
+            "options",
+            String::from("uwu"),
+            None,
+            ValueSetPolicy::Prepend,
+        );
+        assert_eq!(
+            entry.options,
+            vec![
+                BLSValue::Value(String::from("uwu")),
+                BLSValue::Value(String::from("wtf"))
+            ]
+        );
+    }
 }
